@@ -82,10 +82,11 @@ struct input_block
       struct
 	{
 	  FILE *file;		/* input file handle */
+	  boolean end;		/* true if peek has seen EOF */
+	  boolean close;	/* true if we should close file on pop */
 	  const char *name;	/* name of PREVIOUS input file */
-	  int lineno;		/* current line number for do */
-	  /* Yet another attack of "The curse of global variables" (sic) */
-	  int out_lineno;	/* current output line number do */
+	  int lineno;		/* current line of previous file */
+	  int out_lineno;	/* current output line of previous file */
 	  boolean advance_line;	/* start_of_input_line from next_char () */
 	}
       u_f;
@@ -140,24 +141,32 @@ STRING ecomm;
 
 #ifdef ENABLE_CHANGEWORD
 
-#define DEFAULT_WORD_REGEXP "[_a-zA-Z][_a-zA-Z0-9]*"
+# define DEFAULT_WORD_REGEXP "[_a-zA-Z][_a-zA-Z0-9]*"
 
 static char *word_start;
 static struct re_pattern_buffer word_regexp;
 static int default_word_regexp;
 static struct re_registers regs;
 
-#endif /* ENABLE_CHANGEWORD */
+#else /* ! ENABLE_CHANGEWORD */
+# define default_word_regexp 1
+#endif /* ! ENABLE_CHANGEWORD */
+
+#ifdef DEBUG_INPUT
+static const char *token_type_string (token_type);
+#endif
 
 
-/*-------------------------------------------------------------------------.
-| push_file () pushes an input file on the input stack, saving the current |
-| file name and line number.  If next is non-NULL, this push invalidates a |
-| call to push_string_init (), whose storage are consequentely released.   |
-`-------------------------------------------------------------------------*/
+/*-------------------------------------------------------------------.
+| push_file () pushes an input file on the input stack, saving the   |
+| current file name and line number.  If next is non-NULL, this push |
+| invalidates a call to push_string_init (), whose storage is        |
+| consequently released.  If CLOSE, then close FP after EOF is       |
+| detected.                                                          |
+`-------------------------------------------------------------------*/
 
 void
-push_file (FILE *fp, const char *title)
+push_file (FILE *fp, const char *title, boolean close)
 {
   input_block *i;
 
@@ -174,6 +183,8 @@ push_file (FILE *fp, const char *title)
 				     sizeof (struct input_block));
   i->type = INPUT_FILE;
 
+  i->u.u_f.end = FALSE;
+  i->u.u_f.close = close;
   i->u.u_f.name = current_file;
   i->u.u_f.lineno = current_line;
   i->u.u_f.out_lineno = output_current_line;
@@ -229,7 +240,7 @@ push_string_init (void)
     }
 
   next = (input_block *) obstack_alloc (current_input,
-				        sizeof (struct input_block));
+					sizeof (struct input_block));
   next->type = INPUT_STRING;
   return current_input;
 }
@@ -276,8 +287,9 @@ push_string_finish (void)
 void
 push_wrapup (const char *s)
 {
-  input_block *i = (input_block *) obstack_alloc (wrapup_stack,
-						  sizeof (struct input_block));
+  input_block *i;
+  i = (input_block *) obstack_alloc (wrapup_stack,
+				     sizeof (struct input_block));
   i->prev = wsp;
   i->type = INPUT_STRING;
   i->u.u_s.string = obstack_copy0 (wrapup_stack, s, strlen (s));
@@ -304,16 +316,45 @@ pop_input (void)
 
     case INPUT_FILE:
       if (debug_level & DEBUG_TRACE_INPUT)
-	DEBUG_MESSAGE2 ("input reverted to %s, line %d",
-			isp->u.u_f.name, isp->u.u_f.lineno);
+	{
+	  if (isp->u.u_f.lineno)
+	    DEBUG_MESSAGE2 ("input reverted to %s, line %d",
+			    isp->u.u_f.name, isp->u.u_f.lineno);
+	  else
+	    DEBUG_MESSAGE ("input exhausted");
+	}
 
-      fclose (isp->u.u_f.file);
+      if (ferror (isp->u.u_f.file))
+	{
+	  M4ERROR ((warning_status, 0, "read error"));
+	  fclose (isp->u.u_f.file);
+	  retcode = EXIT_FAILURE;
+	}
+      else if (isp->u.u_f.close && fclose (isp->u.u_f.file) == EOF)
+	{
+	  M4ERROR ((warning_status, errno, "error reading file"));
+	  retcode = EXIT_FAILURE;
+	}
       current_file = isp->u.u_f.name;
       current_line = isp->u.u_f.lineno;
       output_current_line = isp->u.u_f.out_lineno;
       start_of_input_line = isp->u.u_f.advance_line;
-      if (tmp != NULL)
-	output_current_line = -1;
+      if (tmp == NULL)
+	{
+	  /* We have exhausted the current input stack.  However,
+	     freeing the obstack now is a bad idea, since if we are in
+	     the middle of a quote, comment, dnl, or argument
+	     collection, there is still a pointer to the former
+	     current_file that we must not invalidate until after the
+	     warning message has been issued.  Setting next to a
+	     non-string is safe in this case, because the only place
+	     more input could come from is another push_file or
+	     pop_wrapup, both of which then free the input_block.  */
+	  next = isp;
+	  isp = NULL;
+	  return;
+	}
+      output_current_line = -1;
       break;
 
     default:
@@ -336,13 +377,14 @@ pop_input (void)
 boolean
 pop_wrapup (void)
 {
+  next = NULL;
   obstack_free (current_input, NULL);
-  xfree (current_input);
+  free (current_input);
 
   if (wsp == NULL)
     {
       obstack_free (wrapup_stack, NULL);
-      xfree (wrapup_stack);
+      free (wrapup_stack);
       return FALSE;
     }
 
@@ -383,10 +425,10 @@ init_macro_token (token_data *td)
 | input stack.								  |
 `------------------------------------------------------------------------*/
 
-int
+static int
 peek_input (void)
 {
-  register int ch;
+  int ch;
 
   while (1)
     {
@@ -396,7 +438,7 @@ peek_input (void)
       switch (isp->type)
 	{
 	case INPUT_STRING:
-	  ch = isp->u.u_s.string[0];
+	  ch = to_uchar (isp->u.u_s.string[0]);
 	  if (ch != '\0')
 	    return ch;
 	  break;
@@ -408,6 +450,7 @@ peek_input (void)
 	      ungetc (ch, isp->u.u_f.file);
 	      return ch;
 	    }
+	  isp->u.u_f.end = TRUE;
 	  break;
 
 	case INPUT_MACRO:
@@ -418,8 +461,12 @@ peek_input (void)
 		    "INTERNAL ERROR: input stack botch in peek_input ()"));
 	  abort ();
 	}
-      /* End of input source --- pop one level.  */
-      pop_input ();
+      /* End of current input source --- pop one level if another
+	 level still exists.  */
+      if (isp->prev != NULL)
+	pop_input ();
+      else
+	return CHAR_EOF;
     }
 }
 
@@ -435,13 +482,13 @@ peek_input (void)
 
 #define next_char() \
   (isp && isp->type == INPUT_STRING && isp->u.u_s.string[0]		\
-   ? *isp->u.u_s.string++						\
+   ? to_uchar (*isp->u.u_s.string++)					\
    : next_char_1 ())
 
 static int
 next_char_1 (void)
 {
-  register int ch;
+  int ch;
 
   if (start_of_input_line)
     {
@@ -457,13 +504,16 @@ next_char_1 (void)
       switch (isp->type)
 	{
 	case INPUT_STRING:
-	  ch = *isp->u.u_s.string++;
+	  ch = to_uchar (*isp->u.u_s.string++);
 	  if (ch != '\0')
 	    return ch;
 	  break;
 
 	case INPUT_FILE:
-	  ch = getc (isp->u.u_f.file);
+	  /* If stdin is a terminal, calling getc after peek_input
+	     already called it would make the user have to hit ^D
+	     twice to quit.  */
+	  ch = isp->u.u_f.end ? EOF : getc (isp->u.u_f.file);
 	  if (ch != EOF)
 	    {
 	      if (ch == '\n')
@@ -497,42 +547,61 @@ void
 skip_line (void)
 {
   int ch;
+  const char *file = current_file;
+  int line = current_line;
 
   while ((ch = next_char ()) != CHAR_EOF && ch != '\n')
     ;
+  if (ch == CHAR_EOF)
+    /* current_file changed to "" if we see CHAR_EOF, use the
+       previous value we stored earlier.  */
+    M4ERROR_AT_LINE ((warning_status, 0, file, line,
+		      "Warning: end of file treated as newline"));
 }
 
 
-/*----------------------------------------------------------------------.
-| This function is for matching a string against a prefix of the input  |
-| stream.  If the string matches the input, the input is discarded,     |
-| otherwise the characters read are pushed back again.  The function is |
-| used only when multicharacter quotes or comment delimiters are used.  |
-`----------------------------------------------------------------------*/
+/*------------------------------------------------------------------.
+| This function is for matching a string against a prefix of the    |
+| input stream.  If the string matches the input and consume is     |
+| TRUE, the input is discarded; otherwise any characters read are   |
+| pushed back again.  The function is used only when multicharacter |
+| quotes or comment delimiters are used.                            |
+`------------------------------------------------------------------*/
 
-static int
-match_input (const char *s)
+static boolean
+match_input (const char *s, boolean consume)
 {
   int n;			/* number of characters matched */
   int ch;			/* input character */
   const char *t;
+  boolean result = FALSE;
 
   ch = peek_input ();
-  if (ch != *s)
-    return 0;			/* fail */
-  (void) next_char ();
+  if (ch != to_uchar (*s))
+    return FALSE;			/* fail */
 
   if (s[1] == '\0')
-    return 1;			/* short match */
-
-  for (n = 1, t = s++; (ch = peek_input ()) == *s++; n++)
     {
-      (void) next_char ();
-      if (*s == '\0')		/* long match */
-	return 1;
+      if (consume)
+	(void) next_char ();
+      return TRUE;			/* short match */
     }
 
-  /* Failed, push back input.  */
+  (void) next_char ();
+  for (n = 1, t = s++; (ch = peek_input ()) == to_uchar (*s++); )
+    {
+      (void) next_char ();
+      n++;
+      if (*s == '\0')		/* long match */
+	{
+	  if (consume)
+	    return TRUE;
+	  result = TRUE;
+	  break;
+	}
+    }
+
+  /* Failed or shouldn't consume, push back input.  */
   {
     struct obstack *h = push_string_init ();
 
@@ -540,21 +609,23 @@ match_input (const char *s)
     obstack_grow (h, t, n);
   }
   push_string_finish ();
-  return 0;
+  return result;
 }
 
-/*------------------------------------------------------------------------.
-| The macro MATCH() is used to match a string against the input.  The	  |
-| first character is handled inline, for speed.  Hopefully, this will not |
-| hurt efficiency too much when single character quotes and comment	  |
-| delimiters are used.							  |
-`------------------------------------------------------------------------*/
+/*--------------------------------------------------------------------.
+| The macro MATCH() is used to match a string S against the input.    |
+| The first character is handled inline, for speed.  Hopefully, this  |
+| will not hurt efficiency too much when single character quotes and  |
+| comment delimiters are used.  If CONSUME, then CH is the result of  |
+| next_char, and a successful match will discard the matched string.  |
+| Otherwise, CH is the result of peek_char, and the input stream is   |
+| effectively unchanged.                                              |
+`--------------------------------------------------------------------*/
 
-#define MATCH(ch, s) \
-  ((s)[0] == (ch) \
-   && (ch) != '\0' \
-   && ((s)[1] == '\0' \
-       || (match_input ((s) + 1) ? (ch) = peek_input (), 1 : 0)))
+#define MATCH(ch, s, consume)                                           \
+  (to_uchar ((s)[0]) == (ch)                                            \
+   && (ch) != '\0'                                                      \
+   && ((s)[1] == '\0' || (match_input ((s) + (consume), consume))))
 
 
 /*----------------------------------------------------------.
@@ -564,7 +635,7 @@ match_input (const char *s)
 void
 input_init (void)
 {
-  current_file = "NONE";
+  current_file = "";
   current_line = 0;
 
   obstack_init (&token_stack);
@@ -606,8 +677,8 @@ input_init (void)
 void
 set_quotes (const char *lq, const char *rq)
 {
-  xfree (lquote.string);
-  xfree (rquote.string);
+  free (lquote.string);
+  free (rquote.string);
 
   lquote.string = xstrdup (lq ? lq : DEF_LQUOTE);
   lquote.length = strlen (lquote.string);
@@ -618,8 +689,8 @@ set_quotes (const char *lq, const char *rq)
 void
 set_comment (const char *bc, const char *ec)
 {
-  xfree (bcomm.string);
-  xfree (ecomm.string);
+  free (bcomm.string);
+  free (ecomm.string);
 
   bcomm.string = xstrdup (bc ? bc : DEF_BCOMM);
   bcomm.length = strlen (bcomm.string);
@@ -629,12 +700,12 @@ set_comment (const char *bc, const char *ec)
 
 #ifdef ENABLE_CHANGEWORD
 
-void
+static void
 init_pattern_buffer (struct re_pattern_buffer *buf)
 {
-  buf->translate = 0;
-  buf->fastmap = 0;
-  buf->buffer = 0;
+  buf->translate = NULL;
+  buf->fastmap = NULL;
+  buf->buffer = NULL;
   buf->allocated = 0;
 }
 
@@ -666,7 +737,9 @@ set_word_regexp (const char *regexp)
 
   /* If compilation worked, retry using the word_regexp struct.
      Can't rely on struct assigns working, so redo the compilation.  */
+  regfree (&word_regexp);
   msg = re_compile_pattern (regexp, strlen (regexp), &word_regexp);
+  re_set_registers (&word_regexp, &regs, regs.num_regs, regs.start, regs.end);
 
   if (msg != NULL)
     {
@@ -685,8 +758,7 @@ set_word_regexp (const char *regexp)
   for (i = 1; i < 256; i++)
     {
       test[0] = i;
-      if (re_search (&word_regexp, test, 1, 0, 0, &regs) >= 0)
-	strcat (word_start, test);
+      word_start[i] = re_search (&word_regexp, test, 1, 0, 0, NULL) >= 0;
     }
 }
 
@@ -717,41 +789,52 @@ next_token (token_data *td)
   int startpos;
   char *orig_text = 0;
 #endif
+  const char *file = current_file;
+  int line = current_line;
 
   obstack_free (&token_stack, token_bottom);
   obstack_1grow (&token_stack, '\0');
   token_bottom = obstack_finish (&token_stack);
 
+ /* Can't consume character until after CHAR_MACRO is handled.  */
   ch = peek_input ();
   if (ch == CHAR_EOF)
     {
 #ifdef DEBUG_INPUT
       fprintf (stderr, "next_token -> EOF\n");
 #endif
+      next_char ();
       return TOKEN_EOF;
     }
   if (ch == CHAR_MACRO)
     {
       init_macro_token (td);
-      (void) next_char ();
+      next_char ();
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "next_token -> MACDEF (%s)\n",
+	       find_builtin_by_addr (TOKEN_DATA_FUNC (td))->name);
+#endif
       return TOKEN_MACDEF;
     }
 
-  (void) next_char ();
-  if (MATCH (ch, bcomm.string))
+  next_char (); /* Consume character we already peeked at.  */
+  if (MATCH (ch, bcomm.string, TRUE))
     {
       obstack_grow (&token_stack, bcomm.string, bcomm.length);
-      while ((ch = next_char ()) != CHAR_EOF && !MATCH (ch, ecomm.string))
+      while ((ch = next_char ()) != CHAR_EOF
+	     && !MATCH (ch, ecomm.string, TRUE))
 	obstack_1grow (&token_stack, ch);
       if (ch != CHAR_EOF)
 	obstack_grow (&token_stack, ecomm.string, ecomm.length);
+      else
+	/* current_file changed to "" if we see CHAR_EOF, use the
+	   previous value we stored earlier.  */
+	M4ERROR_AT_LINE ((EXIT_FAILURE, 0, file, line,
+			  "ERROR: end of file in comment"));
+
       type = TOKEN_STRING;
     }
-#ifdef ENABLE_CHANGEWORD
   else if (default_word_regexp && (isalpha (ch) || ch == '_'))
-#else
-  else if (isalpha (ch) || ch == '_')
-#endif
     {
       obstack_1grow (&token_stack, ch);
       while ((ch = peek_input ()) != CHAR_EOF && (isalnum (ch) || ch == '_'))
@@ -764,11 +847,11 @@ next_token (token_data *td)
 
 #ifdef ENABLE_CHANGEWORD
 
-  else if (!default_word_regexp && strchr (word_start, ch))
+  else if (!default_word_regexp && word_start[ch])
     {
       obstack_1grow (&token_stack, ch);
       while (1)
-        {
+	{
 	  ch = peek_input ();
 	  if (ch == CHAR_EOF)
 	    break;
@@ -800,9 +883,23 @@ next_token (token_data *td)
 
 #endif /* ENABLE_CHANGEWORD */
 
-  else if (!MATCH (ch, lquote.string))
+  else if (!MATCH (ch, lquote.string, TRUE))
     {
-      type = TOKEN_SIMPLE;
+      switch (ch)
+	{
+	case '(':
+	  type = TOKEN_OPEN;
+	  break;
+	case ',':
+	  type = TOKEN_COMMA;
+	  break;
+	case ')':
+	  type = TOKEN_CLOSE;
+	  break;
+	default:
+	  type = TOKEN_SIMPLE;
+	  break;
+	}
       obstack_1grow (&token_stack, ch);
     }
   else
@@ -812,16 +909,18 @@ next_token (token_data *td)
 	{
 	  ch = next_char ();
 	  if (ch == CHAR_EOF)
-	    M4ERROR ((EXIT_FAILURE, 0,
-		      "ERROR: end of file in string"));
+	    /* current_file changed to "" if we see CHAR_EOF, use
+	       the previous value we stored earlier.  */
+	    M4ERROR_AT_LINE ((EXIT_FAILURE, 0, file, line,
+			      "ERROR: end of file in string"));
 
-	  if (MATCH (ch, rquote.string))
+	  if (MATCH (ch, rquote.string, TRUE))
 	    {
 	      if (--quote_level == 0)
 		break;
 	      obstack_grow (&token_stack, rquote.string, rquote.length);
 	    }
-	  else if (MATCH (ch, lquote.string))
+	  else if (MATCH (ch, lquote.string, TRUE))
 	    {
 	      quote_level++;
 	      obstack_grow (&token_stack, lquote.string, lquote.length);
@@ -842,20 +941,127 @@ next_token (token_data *td)
   TOKEN_DATA_ORIG_TEXT (td) = orig_text;
 #endif
 #ifdef DEBUG_INPUT
-  fprintf (stderr, "next_token -> %d (%s)\n", type, TOKEN_DATA_TEXT (td));
+  fprintf (stderr, "next_token -> %s (%s)\n",
+	   token_type_string (type), TOKEN_DATA_TEXT (td));
 #endif
   return type;
 }
+
+/*-----------------------------------------------.
+| Peek at the next token from the input stream.  |
+`-----------------------------------------------*/
+
+token_type
+peek_token (void)
+{
+  int ch = peek_input ();
+
+  if (ch == CHAR_EOF)
+    {
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> EOF\n");
+#endif
+      return TOKEN_EOF;
+    }
+  if (ch == CHAR_MACRO)
+    {
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> MACDEF\n");
+#endif
+      return TOKEN_MACDEF;
+    }
+
+  if (MATCH (ch, bcomm.string, FALSE))
+    {
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> COMMENT\n");
+#endif
+      return TOKEN_STRING;
+    }
+
+  if ((default_word_regexp && (isalpha (ch) || ch == '_'))
+#ifdef ENABLE_CHANGEWORD
+      || (! default_word_regexp && word_start[ch])
+#endif /* ENABLE_CHANGEWORD */
+      )
+    {
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> WORD\n");
+#endif
+      return TOKEN_WORD;
+    }
+
+  if (MATCH (ch, lquote.string, FALSE))
+    {
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> QUOTE\n");
+#endif
+      return TOKEN_STRING;
+    }
+
+  switch (ch)
+    {
+    case '(':
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> OPEN\n");
+#endif
+      return TOKEN_OPEN;
+    case ',':
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> COMMA\n");
+#endif
+      return TOKEN_COMMA;
+    case ')':
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> CLOSE\n");
+#endif
+      return TOKEN_CLOSE;
+    default:
+#ifdef DEBUG_INPUT
+      fprintf (stderr, "peek_token -> SIMPLE\n");
+#endif
+      return TOKEN_SIMPLE;
+    }
+}
 
 
 #ifdef DEBUG_INPUT
 
+static const char *
+token_type_string (token_type t)
+{
+ switch (t)
+    {				/* TOKSW */
+    case TOKEN_EOF:
+      return "EOF";
+    case TOKEN_STRING:
+      return "STRING";
+    case TOKEN_WORD:
+      return "WORD";
+    case TOKEN_OPEN:
+      return "OPEN";
+    case TOKEN_COMMA:
+      return "COMMA";
+    case TOKEN_CLOSE:
+      return "CLOSE";
+    case TOKEN_SIMPLE:
+      return "SIMPLE";
+    case TOKEN_MACDEF:
+      return "MACDEF";
+    default:
+      abort ();
+    }
+ }
+
 static void
 print_token (const char *s, token_type t, token_data *td)
 {
   fprintf (stderr, "%s: ", s);
   switch (t)
     {				/* TOKSW */
+    case TOKEN_OPEN:
+    case TOKEN_COMMA:
+    case TOKEN_CLOSE:
     case TOKEN_SIMPLE:
       fprintf (stderr, "char:");
       break;
