@@ -1,32 +1,31 @@
 /* GNU m4 -- A simple macro processor
 
-   Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 2004 Free
+   Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 2004, 2005, 2006 Free
    Software Foundation, Inc.
-  
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
-  
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-  
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301  USA
 */
 
 #include "m4.h"
 
 #include <getopt.h>
-#include <sys/signal.h>
+#include <limits.h>
+#include <signal.h>
 
-static void usage _((int));
-
-/* Operate interactively (-e).  */
-static int interactive = 0;
+static void usage (int);
 
 /* Enable sync output for /lib/cpp (-s).  */
 int sync_output = 0;
@@ -35,7 +34,7 @@ int sync_output = 0;
 int debug_level = 0;
 
 /* Hash table size (should be a prime) (-Hsize).  */
-int hash_table_size = HASHMAX;
+size_t hash_table_size = HASHMAX;
 
 /* Disable GNU extensions (-G).  */
 int no_gnu_extensions = 0;
@@ -53,28 +52,16 @@ int suppress_warnings = 0;
 int warning_status = 0;
 
 /* Artificial limit for expansion_level in macro.c.  */
-int nesting_limit = 250;
+int nesting_limit = 1024;
 
 #ifdef ENABLE_CHANGEWORD
 /* User provided regexp for describing m4 words.  */
-const char *user_word_regexp = NULL;
+const char *user_word_regexp = "";
 #endif
 
-/* Name of frozen file to digest after initialization.  */
-const char *frozen_file_to_read = NULL;
-
-/* Name of frozen file to produce near completion.  */
-const char *frozen_file_to_write = NULL;
-
 /* The name this program was run with. */
 const char *program_name;
 
-/* If non-zero, display usage information and exit.  */
-static int show_help = 0;
-
-/* If non-zero, print the version on standard output and exit.  */
-static int show_version = 0;
-
 struct macro_definition
 {
   struct macro_definition *next;
@@ -85,18 +72,30 @@ typedef struct macro_definition macro_definition;
 
 /* Error handling functions.  */
 
-/*-------------------------------------------------------------------------.
-| Print source and line reference on standard error, as a prefix for error |
-| messages.  Flush standard output first.				   |
-`-------------------------------------------------------------------------*/
+/*-----------------------.
+| Wrapper around error.  |
+`-----------------------*/
 
 void
-reference_error (void)
+m4_error (int status, int errnum, const char *format, ...)
 {
-  int e = errno;
-  fflush (stdout);
-  fprintf (stderr, "%s:%d: ", current_file, current_line);
-  errno = e;
+  va_list args;
+  va_start (args, format);
+  verror_at_line (status, errnum, current_line ? current_file : NULL,
+		  current_line, format, args);
+}
+
+/*-------------------------------.
+| Wrapper around error_at_line.  |
+`-------------------------------*/
+
+void
+m4_error_at_line (int status, int errnum, const char *file, int line,
+		  const char *format, ...)
+{
+  va_list args;
+  va_start (args, format);
+  verror_at_line (status, errnum, line ? file : NULL, line, format, args);
 }
 
 #ifdef USE_STACKOVF
@@ -109,24 +108,11 @@ static void
 stackovf_handler (void)
 {
   M4ERROR ((EXIT_FAILURE, 0,
-	    "ERROR: Stack overflow.  (Infinite define recursion?)"));
+	    "ERROR: stack overflow.  (Infinite define recursion?)"));
 }
 
 #endif /* USE_STACKOV */
 
-/* Memory allocation.  */
-
-/*------------------------.
-| Failsafe free routine.  |
-`------------------------*/
-
-void
-xfree (voidstar p)
-{
-  if (p != NULL)
-    free (p);
-}
-
 
 /*---------------------------------------------.
 | Print a usage message and exit with STATUS.  |
@@ -141,71 +127,86 @@ usage (int status)
     {
       printf ("Usage: %s [OPTION]... [FILE]...\n", program_name);
       fputs ("\
+Process macros in FILEs.  If no FILE or if FILE is `-', standard input\n\
+is read.\n\
+", stdout);
+      fputs ("\
+\n\
 Mandatory or optional arguments to long options are mandatory or optional\n\
 for short options too.\n\
 \n\
 Operation modes:\n\
       --help                   display this help and exit\n\
       --version                output version information and exit\n\
-  -e, --interactive            unbuffer output, ignore interrupts\n\
   -E, --fatal-warnings         stop execution after first warning\n\
+  -i, --interactive            unbuffer output, ignore interrupts\n\
+  -P, --prefix-builtins        force a `m4_' prefix to all builtins\n\
   -Q, --quiet, --silent        suppress some warnings for builtins\n\
-  -P, --prefix-builtins        force a `m4_' prefix to all builtins\n",
-	     stdout);
+", stdout);
 #ifdef ENABLE_CHANGEWORD
       fputs ("\
-  -W, --word-regexp=REGEXP     use REGEXP for macro name syntax\n",
-	     stdout);
+  -W, --word-regexp=REGEXP     use REGEXP for macro name syntax\n\
+", stdout);
 #endif
       fputs ("\
 \n\
 Preprocessor features:\n\
-  -I, --include=DIRECTORY      search this directory second for includes\n\
-  -D, --define=NAME[=VALUE]    enter NAME has having VALUE, or empty\n\
-  -U, --undefine=NAME          delete builtin NAME\n\
-  -s, --synclines              generate `#line NO \"FILE\"' lines\n",
-	     stdout);
+  -D, --define=NAME[=VALUE]    define NAME as having VALUE, or empty\n\
+  -I, --include=DIRECTORY      append DIRECTORY to include path\n\
+  -s, --synclines              generate `#line NUM \"FILE\"' lines\n\
+  -U, --undefine=NAME          undefine NAME\n\
+", stdout);
       fputs ("\
 \n\
 Limits control:\n\
   -G, --traditional            suppress all GNU extensions\n\
-  -H, --hashsize=PRIME         set symbol lookup hash table size\n\
-  -L, --nesting-limit=NUMBER   change artificial nesting limit\n",
-	     stdout);
+  -H, --hashsize=PRIME         set symbol lookup hash table size [509]\n\
+  -L, --nesting-limit=NUMBER   change artificial nesting limit [1024]\n\
+", stdout);
       fputs ("\
 \n\
 Frozen state files:\n\
   -F, --freeze-state=FILE      produce a frozen state on FILE at end\n\
-  -R, --reload-state=FILE      reload a frozen state from FILE at start\n",
-	     stdout);
+  -R, --reload-state=FILE      reload a frozen state from FILE at start\n\
+", stdout);
       fputs ("\
 \n\
 Debugging:\n\
-  -d, --debug=[FLAGS]          set debug level (no FLAGS implies `aeq')\n\
-  -t, --trace=NAME             trace NAME when it will be defined\n\
+  -d, --debug[=FLAGS]          set debug level (no FLAGS implies `aeq')\n\
+      --debugfile=FILE         redirect debug and trace output\n\
   -l, --arglength=NUM          restrict macro tracing size\n\
-  -o, --error-output=FILE      redirect debug and trace output\n",
-	     stdout);
+  -t, --trace=NAME             trace NAME when it is defined\n\
+", stdout);
       fputs ("\
 \n\
 FLAGS is any of:\n\
-  t   trace for all macro calls, not only traceon'ed\n\
   a   show actual arguments\n\
-  e   show expansion\n\
-  q   quote values as necessary, with a or e flag\n\
   c   show before collect, after collect and after call\n\
-  x   add a unique macro call id, useful with c flag\n\
+  e   show expansion\n\
   f   say current input file name\n\
+  i   show changes in input files\n\
   l   say current input line number\n\
   p   show results of path searches\n\
-  i   show changes in input files\n\
-  V   shorthand for all of the above flags\n",
-	     stdout);
+  q   quote values as necessary, with a or e flag\n\
+  t   trace for all macro calls, not only traceon'ed\n\
+  x   add a unique macro call id, useful with c flag\n\
+  V   shorthand for all of the above flags\n\
+", stdout);
       fputs ("\
 \n\
-If no FILE or if FILE is `-', standard input is read.\n",
-	     stdout);
+If defined, the environment variable `M4PATH' is a colon-separated list\n\
+of directories included after any specified by `-I'.\n\
+", stdout);
+      fputs ("\
+\n\
+Exit status is 0 for success, 1 for failure, 63 for frozen file version\n\
+mismatch, or whatever value was passed to the m4exit macro.\n\
+", stdout);
+      printf ("\nReport bugs to <%s>.\n", PACKAGE_BUGREPORT);
     }
+
+  if (close_stream (stdout) != 0)
+    M4ERROR ((EXIT_FAILURE, errno, "write error"));
   exit (status);
 }
 
@@ -213,41 +214,56 @@ If no FILE or if FILE is `-', standard input is read.\n",
 | Decode options and launch execution.  |
 `--------------------------------------*/
 
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  DEBUGFILE_OPTION = CHAR_MAX + 1,	/* no short opt */
+  DIVERSIONS_OPTION,			/* not quite -N, because of message */
+
+  HELP_OPTION,				/* no short opt */
+  VERSION_OPTION			/* no short opt */
+};
+
 static const struct option long_options[] =
 {
   {"arglength", required_argument, NULL, 'l'},
   {"debug", optional_argument, NULL, 'd'},
-  {"diversions", required_argument, NULL, 'N'},
-  {"error-output", required_argument, NULL, 'o'},
+  {"define", required_argument, NULL, 'D'},
+  {"error-output", required_argument, NULL, 'o'}, /* FIXME: deprecate in 2.0 */
   {"fatal-warnings", no_argument, NULL, 'E'},
   {"freeze-state", required_argument, NULL, 'F'},
   {"hashsize", required_argument, NULL, 'H'},
   {"include", required_argument, NULL, 'I'},
-  {"interactive", no_argument, NULL, 'e'},
+  {"interactive", no_argument, NULL, 'i'},
   {"nesting-limit", required_argument, NULL, 'L'},
   {"prefix-builtins", no_argument, NULL, 'P'},
   {"quiet", no_argument, NULL, 'Q'},
   {"reload-state", required_argument, NULL, 'R'},
   {"silent", no_argument, NULL, 'Q'},
   {"synclines", no_argument, NULL, 's'},
+  {"trace", required_argument, NULL, 't'},
   {"traditional", no_argument, NULL, 'G'},
+  {"undefine", required_argument, NULL, 'U'},
   {"word-regexp", required_argument, NULL, 'W'},
 
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {"debugfile", required_argument, NULL, DEBUGFILE_OPTION},
+  {"diversions", required_argument, NULL, DIVERSIONS_OPTION},
 
-  /* These are somewhat troublesome.  */
-  { "define", required_argument, NULL, 'D' },
-  { "undefine", required_argument, NULL, 'U' },
-  { "trace", required_argument, NULL, 't' },
+  {"help", no_argument, NULL, HELP_OPTION},
+  {"version", no_argument, NULL, VERSION_OPTION},
 
-  { 0, 0, 0, 0 },
+  { NULL, 0, NULL, 0 },
 };
 
+/* Global catchall for any errors that should affect final error status, but
+   where we try to continue execution in the meantime.  */
+int retcode;
+
 #ifdef ENABLE_CHANGEWORD
-#define OPTSTRING "B:D:EF:GH:I:L:N:PQR:S:T:U:W:d::el:o:st:"
+#define OPTSTRING "B:D:EF:GH:I:L:N:PQR:S:T:U:W:d::eil:o:st:"
 #else
-#define OPTSTRING "B:D:EF:GH:I:L:N:PQR:S:T:U:d::el:o:st:"
+#define OPTSTRING "B:D:EF:GH:I:L:N:PQR:S:T:U:d::eil:o:st:"
 #endif
 
 int
@@ -260,8 +276,13 @@ main (int argc, char *const *argv, char *const *envp)
 
   macro_definition *defines;
   FILE *fp;
+  boolean read_stdin = FALSE;
+  boolean interactive = FALSE;
+  const char *frozen_file_to_read = NULL;
+  const char *frozen_file_to_write = NULL;
 
   program_name = argv[0];
+  retcode = EXIT_SUCCESS;
 
   include_init ();
   debug_init ();
@@ -273,22 +294,30 @@ main (int argc, char *const *argv, char *const *envp)
 
   head = tail = NULL;
 
-  while (optchar = getopt_long (argc, argv, OPTSTRING, long_options, NULL),
+  while (optchar = getopt_long (argc, (char **) argv, OPTSTRING,
+				long_options, NULL),
 	 optchar != EOF)
     switch (optchar)
       {
       default:
 	usage (EXIT_FAILURE);
 
-      case 0:
-	break;
-
-      case 'B':			/* compatibility junk */
-      case 'N':
+      case 'B':
       case 'S':
       case 'T':
+	/* Compatibility junk: options that other implementations
+	   support, but which we ignore as no-ops and don't list in
+	   --help.  */
+	error (0, 0, "Warning: `m4 -%c' may be removed in a future release",
+	       optchar);
 	break;
 
+      case 'N':
+      case DIVERSIONS_OPTION:
+	/* -N became an obsolete no-op in 1.4.x.  */
+	error (0, 0, "Warning: `m4 %s' is deprecated",
+	       optchar == 'N' ? "-N" : "--diversions");
+
       case 'D':
       case 'U':
       case 't':
@@ -321,8 +350,8 @@ main (int argc, char *const *argv, char *const *envp)
 	break;
 
       case 'H':
-	hash_table_size = atoi (optarg);
-	if (hash_table_size <= 0)
+	hash_table_size = atol (optarg);
+	if (hash_table_size == 0)
 	  hash_table_size = HASHMAX;
 	break;
 
@@ -362,7 +391,10 @@ main (int argc, char *const *argv, char *const *envp)
 	break;
 
       case 'e':
-	interactive = 1;
+	error (0, 0, "Warning: `m4 -e' is deprecated, use `-i' instead");
+	/* fall through */
+      case 'i':
+	interactive = TRUE;
 	break;
 
       case 'l':
@@ -372,6 +404,11 @@ main (int argc, char *const *argv, char *const *envp)
 	break;
 
       case 'o':
+	/* -o/--error-output are deprecated synonyms of --debugfile,
+	   but don't issue a deprecation warning until autoconf 2.61
+	   or later is more widely established, as such a warning
+	   would interfere with all earlier versions of autoconf.  */
+      case DEBUGFILE_OPTION:
 	if (!debug_set_output (optarg))
 	  error (0, errno, "%s", optarg);
 	break;
@@ -379,17 +416,27 @@ main (int argc, char *const *argv, char *const *envp)
       case 's':
 	sync_output = 1;
 	break;
+
+      case VERSION_OPTION:
+	 printf ("%s\n", PACKAGE_STRING);
+	 fputs ("\
+Copyright (C) 2006 Free Software Foundation, Inc.\n\
+This is free software; see the source for copying conditions.  There is NO\n\
+warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
+\n\
+Written by Rene' Seindal.\n\
+", stdout);
+
+	 if (close_stream (stdout) != 0)
+	    M4ERROR ((EXIT_FAILURE, errno, "write error"));
+	 exit (EXIT_SUCCESS);
+	break;
+
+      case HELP_OPTION:
+	usage (EXIT_SUCCESS);
+	break;
       }
 
-  if (show_version)
-    {
-      printf ("GNU %s %s\n", PRODUCT, VERSION);
-      exit (EXIT_SUCCESS);
-    }
-
-  if (show_help)
-    usage (EXIT_SUCCESS);
-
   defines = head;
 
   /* Do the basic initialisations.  */
@@ -435,12 +482,12 @@ main (int argc, char *const *argv, char *const *envp)
 
 	default:
 	  M4ERROR ((warning_status, 0,
-		    "INTERNAL ERROR: Bad code in deferred arguments"));
+		    "INTERNAL ERROR: bad code in deferred arguments"));
 	  abort ();
 	}
 
       next = defines->next;
-      xfree ((voidstar) defines);
+      free (defines);
       defines = next;
     }
 
@@ -457,24 +504,38 @@ main (int argc, char *const *argv, char *const *envp)
 
   if (optind == argc)
     {
-      push_file (stdin, "stdin");
+      /* No point closing stdin until after wrapped text is
+	 processed.  */
+      push_file (stdin, "stdin", FALSE);
+      read_stdin = TRUE;
       expand_input ();
     }
   else
     for (; optind < argc; optind++)
       {
 	if (strcmp (argv[optind], "-") == 0)
-	  push_file (stdin, "stdin");
+	  {
+	    /* If stdin is a terminal, we want to allow 'm4 - file -'
+	       to read input from stdin twice, like GNU cat.  Besides,
+	       there is no point closing stdin before wrapped text, to
+	       minimize bugs in syscmd called from wrapped text.  */
+	    push_file (stdin, "stdin", FALSE);
+	    read_stdin = TRUE;
+	  }
 	else
 	  {
-	    fp = path_search (argv[optind]);
+	    const char *name;
+	    fp = path_search (argv[optind], &name);
 	    if (fp == NULL)
 	      {
 		error (0, errno, "%s", argv[optind]);
+		/* Set the status to EXIT_FAILURE, even though we
+		   continue to process files after a missing file.  */
+		retcode = EXIT_FAILURE;
 		continue;
 	      }
-	    else
-	      push_file (fp, argv[optind]);
+	    push_file (fp, name, TRUE);
+	    free ((char *) name);
 	  }
 	expand_input ();
       }
@@ -485,6 +546,16 @@ main (int argc, char *const *argv, char *const *envp)
   while (pop_wrapup ())
     expand_input ();
 
+  /* Change debug stream back to stderr, to force flushing the debug
+     stream and detect any errors it might have encountered.  Close
+     stdin if we read from it, to detect any errors.  */
+  debug_set_output (NULL);
+  if (read_stdin && fclose (stdin) == EOF)
+    {
+      M4ERROR ((warning_status, errno, "error reading file"));
+      retcode = EXIT_FAILURE;
+    }
+
   if (frozen_file_to_write)
     produce_frozen_state (frozen_file_to_write);
   else
@@ -493,5 +564,7 @@ main (int argc, char *const *argv, char *const *envp)
       undivert_all ();
     }
 
-  exit (EXIT_SUCCESS);
+  if (close_stream (stdout) != 0)
+    M4ERROR ((EXIT_FAILURE, errno, "write error"));
+  exit (retcode);
 }
